import sys
import chameleontools.StealthGenome as _GenomeFile

class _pipeline:
    
    def _parseInputGenome(self, infile: str) -> tuple[str,str]:
        pass
    
    def __init__(self) -> None:
        pass
    
    
        